/*
   Copyright (C) 1994-2001 Digitool, Inc
   This file is part of Opensourced MCL.

   Opensourced MCL is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   Opensourced MCL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

	
/* 01/29/97 gb  GCC wants a __main function.  Don't ask. */
/* 03/26/96 gb  float->unsigned runtime support for MWC */
/* 03/10/96 gb  set_fpscr; use macros for prolog, epilog */
/* 03/04/96 gb  find_interrupt_context */
/* 12/27/95 gb  count_leading_zeros */
/* 12/13/95 gb  zero_cache_lines fencepost, mwc glue */

	include(lisp.s)

	_beginfile
	.endfunc
/* Zero R4 cache lines, starting at address in R3.  Each line is assumed to be 
   R5 bytes wide. */
_exportfn(zero_cache_lines)
	cmpwi cr0,r4,0
	mtctr r4
	beqlr
1:	dcbz 0,r3
	add r3,r3,r5
	bdnz 1b
	blr
_endfn(zero_cache_lines)

/* Flush R4 cache lines, starting at address in R3.  Each line is
   assumed to be R5 bytes wide. */
_exportfn(flush_cache_lines)
	cmpwi cr0,r4,0
	mtctr r4
	beqlr
1:	dcbf 0,r3
	sync
	icbi 0,r3
	isync
	add r3,r3,r5
	bdnz 1b
	blr
_endfn(flush_cache_lines)

_exportfn(current_stack_pointer)
	mr r3,sp
	blr
	b exp
_endfn(current_stack_pointer)
	
_exportfn(setr2)
	mr r2,r3
	blr
_endfn(setr2)
	
_exportfn(count_leading_zeros)
	cntlzw r3,r3
	blr
_endfn(count_leading_zeros)


_exportfn(set_fpscr)
	stwu sp,-16(sp)
	stw r3,12(sp)
	lfd f0,8(sp)
	mtfsf 0xff,f0
	la sp,16(sp)
	blr
_endfn(set_fpscr)

/* The Linux kernel is constantly enabling and disabling the FPU and enabling
   FPU exceptions.  We can't touch the FPU without turning off the FPSCR[FEX]
   bit and we can't turn off the FPSCR[FEX] bit without touching the FPU.
   Force a distinguished exception, and let the handler for that exception
   zero the fpscr in its exception context.
*/
_exportfn(zero_fpscr)
	uuo_zero_fpscr
	blr
_endfn(zero_fpscr)
	
	
_exportfn(save_fp_context)
	subi r4,r3,8
	stfdu f0,8(r4)
	stfdu f1,8(r4)
	stfdu f2,8(r4)
	stfdu f3,8(r4)
	stfdu f4,8(r4)
	stfdu f5,8(r4)
	stfdu f6,8(r4)
	stfdu f7,8(r4)
	stfdu f8,8(r4)
	stfdu f9,8(r4)
	stfdu f10,8(r4)
	stfdu f11,8(r4)
	stfdu f12,8(r4)
	stfdu f13,8(r4)
	stfdu f14,8(r4)
	stfdu f15,8(r4)
	stfdu f16,8(r4)
	stfdu f17,8(r4)
	stfdu f18,8(r4)
	stfdu f19,8(r4)
	stfdu f20,8(r4)
	stfdu f21,8(r4)
	stfdu f22,8(r4)
	stfdu f23,8(r4)
	stfdu f24,8(r4)
	stfdu f25,8(r4)
	stfdu f26,8(r4)
	stfdu f27,8(r4)
	stfdu f28,8(r4)
	stfdu f29,8(r4)
	stfdu f30,8(r4)
	stfdu f31,8(r4)
	mffs f0
	stfd f0,8(r4)
	lfd f0,0(r3)
	blr
_endfn(save_fp_context)

_exportfn(restore_fp_context)
	mr r4,r3
	lfdu f1,8(r4)
	lfdu f2,8(r4)
	lfdu f3,8(r4)
	lfdu f4,8(r4)
	lfdu f5,8(r4)
	lfdu f6,8(r4)
	lfdu f7,8(r4)
	lfdu f8,8(r4)
	lfdu f9,8(r4)
	lfdu f10,8(r4)
	lfdu f11,8(r4)
	lfdu f12,8(r4)
	lfdu f13,8(r4)
	lfdu f14,8(r4)
	lfdu f15,8(r4)
	lfdu f16,8(r4)
	lfdu f17,8(r4)
	lfdu f18,8(r4)
	lfdu f19,8(r4)
	lfdu f20,8(r4)
	lfdu f21,8(r4)
	lfdu f22,8(r4)
	lfdu f23,8(r4)
	lfdu f24,8(r4)
	lfdu f25,8(r4)
	lfdu f26,8(r4)
	lfdu f27,8(r4)
	lfdu f28,8(r4)
	lfdu f29,8(r4)
	lfdu f30,8(r4)
	lfdu f31,8(r4)
	lfd f0,8(r4)
	mtfsf 0xff,f0
	lfd f0,0(r3)
	blr
_endfn(restore_fp_context)

_exportfn(getHID0)
	mfspr r3,1008		/* get HID0 */
	blr
_endfn(getHID0)

_exportfn(setHID0)
	mtspr 1008,r3
	blr
_endfn(setHID0)

	/* Defunct entries in the 'kernel imports table' come here.
	   Should never happen. */
_exportfn(defunct_import)
	lis r3,defunct_msg@ha
	addi r3,r3,defunct_msg@l
	tw 28,sp,sp
	blr
defunct_msg:	
	.asciz "defunct import\n"
	.align 2
_endfn(defunct_import)
	
_exportfn(altivec_probe)
	mfspr r3,256		/* 256 = VRSAVE */
	blr
_endfn(altivec_probe)

/*
	Copy all 32 Altivec registers (+ VSCR & VRSAVE) to the buffer
	in r3.  If the buffer's non-NULL, it's aligned and big enough,
	and Altivec is present.
*/
_exportfn(put_altivec_registers)
	cmpwi r3,0
	li r4,0
	beqlr
	STVX 0,3,4 ; la r4,16(r4)
	STVX 1,3,4 ; la r4,16(r4)
	STVX 2,3,4 ; la r4,16(r4)
	STVX 3,3,4 ; la r4,16(r4)
	STVX 4,3,4 ; la r4,16(r4)
	STVX 5,3,4 ; la r4,16(r4)
	STVX 6,3,4 ; la r4,16(r4)
	STVX 7,3,4 ; la r4,16(r4)
	STVX 8,3,4 ; la r4,16(r4)
	STVX 9,3,4 ; la r4,16(r4)
	STVX 10,3,4 ; la r4,16(r4)
	STVX 11,3,4 ; la r4,16(r4)
	STVX 12,3,4 ; la r4,16(r4)
	STVX 13,3,4 ; la r4,16(r4)
	STVX 14,3,4 ; la r4,16(r4)
	STVX 15,3,4 ; la r4,16(r4)
	STVX 16,3,4 ; la r4,16(r4)
	STVX 17,3,4 ; la r4,16(r4)
	STVX 18,3,4 ; la r4,16(r4)
	STVX 19,3,4 ; la r4,16(r4)
	STVX 20,3,4 ; la r4,16(r4)
	STVX 21,3,4 ; la r4,16(r4)
	STVX 22,3,4 ; la r4,16(r4)
	STVX 23,3,4 ; la r4,16(r4)
	STVX 24,3,4 ; la r4,16(r4)
	STVX 25,3,4 ; la r4,16(r4)
	STVX 26,3,4 ; la r4,16(r4)
	STVX 27,3,4 ; la r4,16(r4)
	STVX 28,3,4 ; la r4,16(r4)
	STVX 29,3,4 ; la r4,16(r4)
	STVX 30,3,4 ; la r4,16(r4)
	STVX 31,3,4 ; la r4,16(r4)
	MFVSCR 0
	STVX 0,3,4
	mfspr r5,256
	stw r5,8(r4)
	blr
_endfn(put_altivec_registers)

_exportfn(get_altivec_registers)
	cmpwi r3,0
	li r4,32*16
	beqlr
	LVX 0,3,4
	MTVSCR 0
	lwz r5,8(r4)
	mtspr 256,r5
	la r4,-16(r4) ; LVX 31,3,4
	la r4,-16(r4) ; LVX 30,3,4
	la r4,-16(r4) ; LVX 29,3,4
	la r4,-16(r4) ; LVX 28,3,4
	la r4,-16(r4) ; LVX 27,3,4
	la r4,-16(r4) ; LVX 26,3,4
	la r4,-16(r4) ; LVX 25,3,4
	la r4,-16(r4) ; LVX 24,3,4
	la r4,-16(r4) ; LVX 23,3,4
	la r4,-16(r4) ; LVX 22,3,4
	la r4,-16(r4) ; LVX 21,3,4
	la r4,-16(r4) ; LVX 20,3,4
	la r4,-16(r4) ; LVX 19,3,4
	la r4,-16(r4) ; LVX 18,3,4
	la r4,-16(r4) ; LVX 17,3,4
	la r4,-16(r4) ; LVX 16,3,4
	la r4,-16(r4) ; LVX 15,3,4
	la r4,-16(r4) ; LVX 14,3,4
	la r4,-16(r4) ; LVX 13,3,4
	la r4,-16(r4) ; LVX 12,3,4
	la r4,-16(r4) ; LVX 11,3,4
	la r4,-16(r4) ; LVX 10,3,4
	la r4,-16(r4) ; LVX 9,3,4
	la r4,-16(r4) ; LVX 8,3,4
	la r4,-16(r4) ; LVX 7,3,4
	la r4,-16(r4) ; LVX 6,3,4
	la r4,-16(r4) ; LVX 5,3,4
	la r4,-16(r4) ; LVX 4,3,4
	la r4,-16(r4) ; LVX 3,3,4
	la r4,-16(r4) ; LVX 2,3,4
	la r4,-16(r4) ; LVX 1,3,4
	la r4,-16(r4) ; LVX 0,3,4
	blr
_endfn(get_altivec_registers)

	_endfile
