/*
   Copyright (C) 1994-2001 Digitool, Inc
   This file is part of Opensourced MCL.

   Opensourced MCL is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   Opensourced MCL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

	include(lisp.s)

/* Frame needs room for backlink, lr, 18 GPRs, FPSCR, 18 FPRs, 12 VRs, the
	VSCR, and VRSAVE, and must be aligned on a 16-byte boundary */

QT_SAVEr14 = 8
QT_SAVEfpscr = QT_SAVEr14 + ((32-14)*4)
QT_SAVEf14 = QT_SAVEfpscr+8
QT_SAVEpad = QT_SAVEf14+((32-14)*8)
QT_SAVEv20 = QT_SAVEpad+8
QT_SAVEvtemp = QT_SAVEv20+((32-20)*16)
QT_SAVEvrsave = QT_SAVEvtemp+8
QT_FRAMESIZE = QT_SAVEvtemp+16

	.if (QT_SAVEv20 & 0xf)
	.fail QT_SAVEv20
	.endif
	
	.if (QT_FRAMESIZE & 0xf)
	.fail QT_FRAMESIZE
	.endif
	 	
/*
	In general, a frame looks like:
+---------+
| backptr |	back link to previous frame
+---------+
| save lr |	slot for CALLEE to save lr in
+---------+
| save r14|	user function param 'pu'
+---------+
| save r15|	user function param 'pt'
+---------+
| save r16|	user function 'userf'
+---------+
| save r17|	'only'
+---------+
    ...
+---------+
| save r31|
+---------+
| fpscr   |	64 bits
|	  |
+---------+
| save f14|
|	  |
+---------+
   ...
+---------+
| save f31|
|	  |
+---------+
|  pad    |      64 bits
|         |
+---------+
| save v20|	128 bits
|         |
+---------+
   ...
+---------+
| save v31|
|         |
+---------+
| vrtemp  |  xxxx, xxxx, vrsave, vscr
|         |
+---------+
		When 'qt_start' runs, there should be a backlink
		and save_lr slot underneath the top frame.
*/

	_beginfile
	.endfunc
	.globl altivec_present
_exportfn(qt_block)
	andi. r0,sp,8
	mflr r0
	stw r0,4(sp)
	bne 1f
	stwu sp,-QT_FRAMESIZE(sp)
	b 2f
1:	stwu sp,-(QT_FRAMESIZE+8)(sp)
2:	
	stmw r14,QT_SAVEr14(sp)
	lis r15,altivec_present@ha
	lwz r15,altivec_present@l(r15)
	cmpwi r15,0
	la r14,QT_SAVEfpscr-8(sp)
	mffs f0
	stfdu f0,8(r14)
	stfdu f14,8(r14)
	stfdu f15,8(r14)
	stfdu f16,8(r14)
	stfdu f17,8(r14)
	stfdu f18,8(r14)
	stfdu f19,8(r14)
	stfdu f20,8(r14)
	stfdu f21,8(r14)
	stfdu f22,8(r14)
	stfdu f23,8(r14)
	stfdu f24,8(r14)
	stfdu f25,8(r14)
	stfdu f26,8(r14)
	stfdu f27,8(r14)
	stfdu f28,8(r14)
	stfdu f29,8(r14)
	stfdu f30,8(r14)
	stfdu f31,8(r14)
	beq 3f			/* branch if no Altivec */
	la r14,8(r14)
	STVX 20,0,14 ; la r14,16(r14)
	STVX 21,0,14 ; la r14,16(r14)
	STVX 22,0,14 ; la r14,16(r14)
	STVX 23,0,14 ; la r14,16(r14)
	STVX 24,0,14 ; la r14,16(r14)
	STVX 25,0,14 ; la r14,16(r14)
	STVX 26,0,14 ; la r14,16(r14)
	STVX 27,0,14 ; la r14,16(r14)
	STVX 28,0,14 ; la r14,16(r14)
	STVX 29,0,14 ; la r14,16(r14)
	STVX 30,0,14 ; la r14,16(r14)
	STVX 31,0,14 ; la r14,16(r14)
	MFVSCR 31
	STVX 31,0,14
	mfspr r14,256		/* VRSAVE */
	stw r14,QT_SAVEvrsave(sp)
3:
	mtctr r3
	mr r3,sp
	mr sp,r6
	bctrl
	lis r15,altivec_present@ha
	lwz r15,altivec_present@l(r15)
	cmpwi r15,0
	la r14,8+(18*4)-8(sp)
	lfdu f0,8(r14)
	mtfsf 0xff,f0
	lfdu f14,8(r14)
	lfdu f15,8(r14)
	lfdu f16,8(r14)
	lfdu f17,8(r14)
	lfdu f18,8(r14)
	lfdu f19,8(r14)
	lfdu f20,8(r14)
	lfdu f21,8(r14)
	lfdu f22,8(r14)
	lfdu f23,8(r14)
	lfdu f24,8(r14)
	lfdu f25,8(r14)
	lfdu f26,8(r14)
	lfdu f27,8(r14)
	lfdu f28,8(r14)
	lfdu f29,8(r14)
	lfdu f30,8(r14)
	lfdu f31,8(r14)
	beq 4f
	la r14,8(r14)
	LVX 20,0,14 ; la r14,16(r14)
	LVX 21,0,14 ; la r14,16(r14)
	LVX 22,0,14 ; la r14,16(r14)
	LVX 23,0,14 ; la r14,16(r14)
	LVX 24,0,14 ; la r14,16(r14)
	LVX 25,0,14 ; la r14,16(r14)
	LVX 26,0,14 ; la r14,16(r14)
	LVX 27,0,14 ; la r14,16(r14)
	LVX 28,0,14 ; la r14,16(r14)
	LVX 29,0,14 ; la r14,16(r14)
	LVX 30,0,14 ; la r14,16(r14)
	LVX 31,0,14 ; la r14,16(r14)
	LVX 0,0,14
	MTVSCR 0
	lwz r14,QT_SAVEvrsave(sp)
	mtspr 256,r14
4:	
	lmw r14,8(sp)
	lwz sp,0(sp)
	lwz r0,4(sp)
	mtlr r0
	blr
_endfn(qt_block)

_exportfn(qt_start)
	mtctr r17		/* Ready to call 'only' */
	mr r3,r14		/* Load up user function pu. */
	mr r4,r15		/* ... user function pt */
	mr r5,r16		/* ... user function userf */
	bctrl			/* Call only(pu,pt,userf) */
	b qt_error		/* should never get here */
_endfn(qt_start)

	_endfile

	